/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "article-db.h"
#include "log.h"
#include "prefs.h" /* for data_dir */
#include "util.h"

//=========================================================================

typedef struct
{
	pan_db db;
	gchar* filename;
	const Server* server;
	const Group* group;
} AHDB;

//=========================================================================

static gchar*
get_filename (const Server *s,
	      const Group  *g)
{
	g_return_val_if_fail (s!=NULL, NULL);
	g_return_val_if_fail (g!=NULL, NULL);

	return g_strdup_printf ("/%s/%s/%s.db", data_dir, s->name, g->name);
}

article_db
ahdb_ref (const Server *s,
	  const Group  *g)
{
	AHDB* ahdb = g_new0 (AHDB, 1);
	ahdb->filename = get_filename ( s, g );
	ahdb->db = pan_db_ref ( ahdb->filename );
	ahdb->server = s;
	ahdb->group = g;

	return (article_db)ahdb;
}

gboolean
ahdb_exists (const Server *s,
	     const Group  *g)
{
	gchar* filename = get_filename (s, g);
	gboolean exists = file_exists (filename);
	g_free (filename);

	return exists;
}

void
ahdb_delete_db_file (const Server *s,
		     const Group  *g)
{
	gchar* filename = get_filename (s, g);
	remove (filename);
	g_free (filename);
}

void
ahdb_unref (article_db db)
{
	AHDB* ahdb = (AHDB*)db;

	g_free (ahdb->filename);
	ahdb->filename = 0;
	pan_db_unref (ahdb->db);
	ahdb->db = 0;

	g_free (ahdb);
}

//=========================================================================

const char *
ahdb_field (const char *flat_article,
	    int         offset)
{
	int i = 0;
	const char *p = flat_article;

	if (offset == 0)
		return p;
	while (*p != '\0') {
		if (*p == '\n')
			i++;
		if (i == offset)
			return (p+1);
		p++;
	}
	return NULL;
}


void
ahdb_save_with_buffer (article_db       db,
                       const Article  * article,
                       gboolean         sync,
                       GString        * buf)
{
	AHDB* ahdb = (AHDB*)db;

	g_string_sprintf ( buf,
		"%d\n"
		"%lu\n"
		"%d\n"
		"%d\n"
		"%d\n"
		"%lu\n"
		"%s\n"	   

		"%s\n"
		"%s\n"
		"%s\n"
		"%s\n"
		"%s\n"
		"%s\n"

		"%d\n",

		(int) article->date,
		(unsigned long)(article->state & ~STATE_NEW),
		article->linecount,
		article->parts,
		article->part,
		article->number,
		(article->filename ? article->filename : ""),

		article->subject,
		article->author,
		(article->references ? article->references : ""),
		(article->newsgroups ? article->newsgroups : ""),
		(article->reply_to ? article->reply_to : ""),
		(article->followup_to ? article->followup_to : ""),

		article->crosspost_qty);

	pan_db_put_value_str (ahdb->db, article->message_id, buf->str);

	if (sync)
		ahdb_sync (db);
}

void
ahdb_save (article_db        db,
	   const Article   * article,
	   gboolean          sync)
{
	GString * str = g_string_sized_new (1024);
	ahdb_save_with_buffer (db, article, sync, str);
	g_string_free (str, TRUE);
}

void
ahdb_save_all (article_db         db,
	       Article         ** articles,
	       int                article_qty,
	       GFunc              gfunc,
	       gpointer           user_data)
{
	int i;
	GString * str = g_string_sized_new (1024);
	for (i=0; i<article_qty; ++i)
	{
		if (gfunc)
			(*gfunc)(articles[i],user_data);

		ahdb_save_with_buffer (db, articles[i], FALSE, str);
	}

	/* cleanup */
	g_string_free (str, TRUE);
	ahdb_sync (db);
}

void
ahdb_sync (article_db db)
{
	AHDB* ahdb = (AHDB*)db;
	pan_db_sync ( ahdb->db );
}

//=========================================================================

static Article*
article_deserialize (const Group * group,
		     const char  * message_id,
		     const char  * db_text)
{
	Article* article;
	const char *pch;
	if (!group)
		return NULL;

	/* break up the db_text */
	g_return_val_if_fail (message_id!=NULL, NULL);
	g_return_val_if_fail (db_text!=NULL, NULL);

	/* build the article data */
	pch = db_text;
	article                 = article_new ();
	article->message_id     = g_strdup (message_id);
	article->date           = get_next_token_int    (pch, '\n', &pch);
	article->state          = get_next_token_int    (pch, '\n', &pch);
	article->linecount      = get_next_token_int    (pch, '\n', &pch);
	article->parts          = get_next_token_int    (pch, '\n', &pch);
	article->part           = get_next_token_int    (pch, '\n', &pch);
	article->number         = get_next_token_ulong  (pch, '\n', &pch);
	article->filename	= get_next_token_str    (pch, '\n', &pch);
	article->subject        = get_next_token_str    (pch, '\n', &pch);
	article->author         = get_next_token_str    (pch, '\n', &pch);
	article->references     = get_next_token_str    (pch, '\n', &pch);
	article->newsgroups     = get_next_token_str    (pch, '\n', &pch);
	article->reply_to       = get_next_token_str    (pch, '\n', &pch);
	article->followup_to    = get_next_token_str    (pch, '\n', &pch);
	article->crosspost_qty  = get_next_token_int    (pch, '\n', &pch);
	article->threads        = NULL;

	/* some simple error-checking */
	if (!is_nonempty_string(article->message_id)
		|| (article->number==0)
		|| !is_nonempty_string(article->subject)
		|| !is_nonempty_string(article->author))
	{
		g_message (_("It appears that article '%s' is corrupt; "
		             "skipping.  If you get an excessive number "
			     "of these, try Group|Empty Group from the main "
			     "menu to flush out the corrupt headers."),
			(is_nonempty_string(article->message_id)
				? article->message_id
				: "(Unknown ID)"));
		pan_object_unref (PAN_OBJECT(article));
		article = NULL;
	}

	return article;
}

Article*
ahdb_get (article_db  db,
	  const char *message_id )
{
	AHDB* ahdb = (AHDB*)db;
	gchar* s = pan_db_get_value_str (ahdb->db, message_id);
	Article* a = article_deserialize (ahdb->group, message_id, s);
	g_free (s);
	return a;
}

typedef struct
{
	AHDB* ahdb;
	GPtrArray* articles;
	GPtrArray* delete_keys;
	GFunc gfunc;
	gpointer user_data;
	unsigned long first;
        int expire_total_qty;
	int expire_read_qty;
}
GetArticlesStruct;

static void
ahdb_get_all_dbfunc (const char *key,
		     const char *val,
		     gpointer    user_data)
{
	GetArticlesStruct* gas = (GetArticlesStruct*) user_data;
	Article* a = article_deserialize (gas->ahdb->group, key, val);

	if (a != NULL)
	{
		/* Expire if it's not on server anymore */
		if (a->number < gas->first)
		{
			gas->expire_total_qty++;
			if (article_flag_on (a, STATE_READ))
				gas->expire_read_qty++;

			/* steal a's message_id string */
			g_ptr_array_add (gas->delete_keys, a->message_id);
			a->message_id = NULL;
			pan_object_unref (PAN_OBJECT(a));
		}
		else
		{
			g_ptr_array_add (gas->articles, a);
			if (gas->gfunc)
				(*gas->gfunc)(a, gas->user_data);
		}
	}
}

GPtrArray*
ahdb_get_all (article_db db,
	      GFunc      gfunc,
	      gpointer   user_data)
{
	AHDB* ahdb = (AHDB*)db;
	GetArticlesStruct gas;
	GPtrArray * articles;
	int i;

	/* reserve pointers for the array */
	articles = g_ptr_array_new ();
	pan_g_ptr_array_reserve (articles, group_get_attrib_i (ahdb->server, ahdb->group, "Total"));

	/* load the articles... */
	gas.articles = articles;
	gas.ahdb = ahdb;
	gas.gfunc = gfunc;
	gas.user_data = user_data;
	gas.expire_read_qty = 0;
	gas.expire_total_qty = 0;
	gas.delete_keys = g_ptr_array_new ();
	gas.first = group_get_attrib_i (ahdb->server, ahdb->group, "First");
	pan_db_foreach (ahdb->db, ahdb_get_all_dbfunc, (gpointer)&gas);

	if (ahdb->group->type == USER_FOLDER)
		return gas.articles;

	/* delete expired articles.  pan_db_foreach locks the db,
	   so we have to do it here instead of in ahdb_get_all_dbfunc. */
	for (i=0; i<gas.delete_keys->len; ++i) {
		char * key = (char*) g_ptr_array_index (gas.delete_keys, i);
		pan_db_erase (gas.ahdb->db, key);
		g_free (key);
	}
	g_ptr_array_free (gas.delete_keys, TRUE);

	group_add_attrib_i (ahdb->server, ahdb->group, "Read", -gas.expire_read_qty);
	group_add_attrib_i (ahdb->server, ahdb->group, "Total", -gas.expire_total_qty);

	log_add_va (_("Expired %d articles (%d were marked read) in group %s"),
		gas.expire_total_qty,
		gas.expire_read_qty,
		ahdb->group->name );

	return gas.articles;
}

//=========================================================================

static void
ahdb_length_dbfunc (const char * key,
		    const char * val,
		    gpointer     user_data)
{
	++(*((int*)user_data));
}

int
ahdb_length (article_db db)
{
	int i = 0;
	pan_db_foreach (((AHDB*)db)->db, ahdb_length_dbfunc, (gpointer)&i);
	return i;
}

//=========================================================================


void
ahdb_erase (article_db          db,
	    const Article     * article,
	    const char        * message_id)
{
	AHDB* ahdb = (AHDB*)db;
	if (article)
		pan_db_erase (ahdb->db, article->message_id);
	else if (message_id)
		pan_db_erase (ahdb->db, message_id);
}

void
ahdb_erase_all (article_db db)
{
	AHDB* ahdb = (AHDB*)db;
	pan_db_erase_all ( ahdb->db );
}

