/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>
#include <string.h>

#include <glib.h>
#include <gnome.h>

#include "acache.h"
#include "debug.h"
#include "server.h"
#include "folderlist.h"
#include "group.h"
#include "gui.h"
#include "article.h"
#include "articlelist.h"
#include "util.h"

/***
****
***/

PanCallback *article_deleted = NULL;

/***
****
***/

static void
article_update_impl (article_db db,
		     const Article * article,
		     gboolean sync,
		     gboolean update_ui,
		     GString * buf);

static void
article_add_flag_impl (article_db db,
		       Article *article,
		       int flag,
		       gboolean sync,
		       gboolean update_ui,
		       GString *buf);

static void
article_remove_flag_impl (const article_db db,
			  Article* article,
			  int flag,
			  gboolean sync,
			  gboolean update_ui,
			  GString *buf);

/***
****
***/


void
article_init (void)
{
	article_deleted = pan_callback_new ();
}

void
article_isolate (Article* article)
{
	GSList *l = NULL;
	Article *parent = article->parent;

	if (parent)
		parent->threads = g_slist_remove (parent->threads, article);

	for (l=article->threads; l!=NULL; l=l->next) {
		Article *child = ARTICLE(l->data);
		child->parent = parent;
		if (parent)
			parent->threads = g_slist_append (parent->threads, child);
	}

	g_slist_free (article->threads);
	article->threads = NULL;
	article->parent = NULL;
}

/**
***
**/

void
article_destructor (PanObject* o)
{
	Article* article = ARTICLE(o);
        g_return_if_fail (article!=NULL);

	/* clean up article parts */
	debug (DEBUG_PAN_OBJECT, "article_destructor: %p", article);
	g_free (article->subject);
	g_free (article->author);
	g_free (article->message_id);
	g_free (article->references);
	g_free (article->newsgroups);
	g_free (article->reply_to);
	g_free (article->followup_to);
	g_free (article->filename);
	g_slist_free (article->threads);

	/* clean up parent */
	pan_object_destructor (o);
}

void
article_constructor (Article* article)
{
	g_return_if_fail (article!=NULL);

	/* init the parent class */
	pan_object_constructor (PAN_OBJECT(article), article_destructor);

	/* init the article bits */
	debug (DEBUG_PAN_OBJECT, "article_constructor: %p", article);
	article->number = 0;
	article->subject = NULL;
	article->author = NULL;
	article->message_id = NULL;
	article->references = NULL;
	article->date = (time_t)0;
	article->linecount = 0;
	article->reply_to = NULL;
	article->followup_to = NULL;
	article->newsgroups = NULL;
	article->parts = 0;
	article->part = 0;
	article->filename = NULL;
	article->parent = NULL;
	article->threads = NULL;
	article->unread_children = 0;
	article->crosspost_qty = 0;
	article->self_passes_filter = FALSE;
	article->tree_passes_filter = FALSE;
}

Article*
article_new (void)
{
	Article* article = g_new (Article, 1);
	article_constructor (article);
	return article;
}

/***
****
***/

/**
 * @pararm article_db
 * @param article article to be marked as read
 * @param marked pointer to an int to be incremented for each article marked read
 *
 * If an article is a multipart article, then the multipart articles are also
 * marked as read if they're not read already.
 */
static void
article_mark_read_impl (article_db db,
			Article* article,
			int* marked,
                        GString* buf)
{
	if (!article)
		return;

	if (!article_flag_on (article, STATE_READ)) {
		article_add_flag_impl (db, article, STATE_READ, FALSE, FALSE, buf);
		articlelist_article_read_changed (article);
		++*marked;
	}

	/* special case: mark ALL of a multipart read */
	if (article->part>0 && article->parts>1 && article->threads!=NULL) {
		GSList *l;
		for (l=article->threads; l!=NULL; l=l->next) {
			Article *child = ARTICLE(l->data);
			article_mark_read_impl (db, child, marked, buf);
		}
	}
}


/**
 * @param server
 * @param group
 * @param article
 * @param sync database
 *
 * This marks an article as read as described in article_mark_read_impl.
 *
 * After being marked, then if sync is true, the database is synchronized.
 *
 * If the read state of more than one article has changed, the group attributes
 * are updated.
 */
void
article_mark_read (const Server* server,
		   const Group* group,
		   Article* article,
		   gboolean sync)
{
	int marked = 0;
	article_db db = NULL;
	GString * buf = g_string_new (NULL);
	
	db = ahdb_ref (server, group);
	article_mark_read_impl (db, article, &marked, buf);

	if (sync)
		ahdb_sync (db);
	if (marked)
		group_add_attrib_i (server, group, "Read", marked);

	ahdb_unref (db);
	g_string_free (buf, TRUE);
}

/**
 * @param server
 * @param group
 * @param article
 * @param sync database
 *
 * This marks a list of articles as read as described in article_mark_read_impl.
 *
 * After being marked, then if sync is true, the database is synchronized.
 *
 * If the read state of more than one article has changed, the group attributes
 * are updated when done.
 */
void
articles_mark_read (const Server* server,
		    const Group* group,
		    Article** articles,
		    int article_qty,
		    gboolean sync)
{
	int marked;
	article_db db;
	int i;
	GString * buf;

	g_return_if_fail (articles!=NULL);
	g_return_if_fail (article_qty>=1);

	if (article_qty>1000) /*FIXME*/
		g_message (_("No, Pan hasn't crashed; marking articles read is just reeeeeal slow. :)  This will be sped up soon..."));

	/* mark the articles as read */
	marked = 0;
       	buf = g_string_new (NULL);
	db = ahdb_ref (server, group);
	for (i=0; i!=article_qty; ++i)
		article_mark_read_impl (db, articles[i], &marked, buf);

	/* if necessary, sync the db and update the "Read" attribute */
	if (sync && marked)
		ahdb_sync (db);
	if (marked)
		group_add_attrib_i (server, group, "Read", marked);

	/* cleanup */
	ahdb_unref (db);
	g_string_free (buf, TRUE);
}


/**
***
**/

/**
 * @pararm article_db
 * @param article article to be marked as read
 * @param marked pointer to an int to be incremented for each article marked unread
 *
 * If an article is a multipart article, then the multipart articles are also
 * marked as unread if they're not unread already.
 */
static void
article_mark_unread_impl (article_db db,
			  Article* article,
			  int *marked,
			  GString *buf)
{
	if (!article)
		return;

	if (article_flag_on (article, STATE_READ)) {
		article_remove_flag_impl (db, article, STATE_READ, FALSE, FALSE, buf);
		articlelist_article_read_changed (article);
		++*marked;
	}

	/* special case: mark ALL of a multipart unread */
	if (article->part>0 && article->parts>1 && article->threads!=NULL) {
		GSList *l;
		for (l=article->threads; l!=NULL; l=l->next) {
			Article *child = ARTICLE(l->data);
			article_mark_unread_impl (db, child, marked, buf);
		}
	}
}

/**
 * @param server
 * @param group
 * @param article
 * @param sync database
 *
 * This marks an article as unread as described in article_mark_unread_impl.
 *
 * After being marked, then if sync is true, the database is synchronized.
 *
 * If the read state of more than one article has changed, the group attributes
 * are updated when done.
 */
void
article_mark_unread (const Server* server,
		     const Group* group,
		     Article* article,
		     gboolean sync)
{
	int marked = 0;
	article_db db = 0;
	GString * buf = g_string_new (NULL);

	db = ahdb_ref (server, group);
	article_mark_unread_impl (db, article, &marked, buf);

	if (sync && marked)
		ahdb_sync (db);
	if (marked)
		group_add_attrib_i (server, group, "Read", -marked);

	ahdb_unref (db);
	g_string_free (buf, TRUE);
}

/**
 * @param server
 * @param group
 * @param article
 * @param sync database
 *
 * Marks a list of articles as unread as described in article_mark_unread_impl.
 *
 * After being marked, then if sync is true, the database is synchronized.
 *
 * If the read state of more than one article has changed, the group attributes
 * are updated when done.
 */
void
articles_mark_unread (const Server* server,
                      const Group* group,
                      Article** articles,
                      int article_qty,
                      gboolean sync )
{
	int marked = 0;
	article_db db = 0;
	GString * buf;
	int i;

	g_return_if_fail (articles!=NULL);
	g_return_if_fail (article_qty>=1);

	db = ahdb_ref (server, group);
	buf = g_string_new (NULL);

	for (i=0; i!=article_qty; ++i)
		article_mark_unread_impl (db, articles[i], &marked, buf);

	if (sync) {
		ahdb_sync (db);
		if (marked)
			group_add_attrib_i (server, group, "Read", -marked);
	}

	g_string_free (buf, TRUE);
	ahdb_unref (db);
}

/***
****
***/

static void
article_delete_impl (article_db db,
                     Article* article,
                     int* read,
                     int* total)
{
	if (!article)
		return;

	/* special case: if a multipart, delete ALL of it */
	if (article->part>0 && article->parts>1 && article->threads!=NULL) {
		GSList *l;
		GSList *tmp = g_slist_copy (article->threads);
		for (l=tmp; l!=NULL; l=l->next) {
			Article *child = ARTICLE(l->data);
			article_delete_impl (db, child, read, total);
		}
		g_slist_free (tmp);
	}

	/* notify listeners before the delete takes place */
	pan_callback_call (article_deleted, article, NULL);

	/* update counts */
	++*total;
	if (article_flag_on (article, STATE_READ))
		++*read;

	/* isolate */
	article_isolate (article);

	/* erase from cache */
	acache_delete (article->message_id);

	/* erase from db */
	ahdb_erase (db, article, NULL);

	/* erase from memory */
	pan_object_unref (PAN_OBJECT(article));
}

void
article_delete (const Server* server,
                const Group* group,
                Article* article,
                gboolean sync)
{
	int total = 0;
	int read = 0;
	article_db db = 0;

	/* sanity checks */
	g_return_if_fail (server!=NULL);
	g_return_if_fail (group!=NULL);
	g_return_if_fail (article!=NULL);

	db = ahdb_ref (server, group);
	article_delete_impl (db, article, &read, &total);
	pan_warn_if_fail (total >= read);

	if (sync)
		ahdb_sync (db);

	group_add_attrib_i (server, group, "Total", -total);
	group_add_attrib_i (server, group, "Read", -read);

	ahdb_unref (db);
}

void
articles_delete (const Server* server,
                 const Group* group,
                 Article** articles,
                 int article_qty,
                 gboolean sync)
{
	int total = 0;
	int read = 0;
	int i;
	article_db db = 0;

	/* sanity checks */
	g_return_if_fail (server!=NULL);
	g_return_if_fail (group!=NULL);
	g_return_if_fail (article_qty>=0);
	g_return_if_fail (article_qty ? articles!=NULL : 1);

	db = ahdb_ref (server, group);
	for (i=0; i!=article_qty; ++i)
		article_delete_impl (db, articles[i], &read, &total);

	if (sync)
		ahdb_sync (db);
	if (total)
		group_add_attrib_i (server, group, "Total", -total);
	if (read)
		group_add_attrib_i (server, group, "Read", -read);

	ahdb_unref (db);
}

/***
****
***/

static void
article_add_flag_impl (article_db db,
                       Article *article,
                       int flag,
                       gboolean sync,
                       gboolean update_ui,
		       GString *buf)
{
	g_return_if_fail (article!=NULL);

	article->state |= flag;
	article_update_impl (db, article, sync, update_ui, buf);
}

void
article_add_flag (const Server *server,
                  const Group *group,
                  Article* article,
                  int flag,
                  gboolean sync)
{
	GString * buf = g_string_new (NULL);
	article_db db = ahdb_ref (server, group);

	article_add_flag_impl (db, article, flag, sync, TRUE, buf);

	ahdb_unref (db);
	g_string_free (buf, TRUE);
}


/***
****
***/

static void
article_remove_flag_impl (const article_db db,
                          Article* article,
                          int flag,
                          gboolean sync,
                          gboolean update_ui,
			  GString * buf)
{
	g_return_if_fail (article!=NULL);

	article->state &= ~flag;
	article_update_impl (db, article, sync, update_ui, buf);
}

void
article_remove_flag (const Server *server,
                     const Group *group,
                     Article* article,
                     int flag,
                     gboolean sync)
{
	GString * buf = g_string_new (NULL);
	article_db db = ahdb_ref (server, group);

	article_remove_flag_impl (db, article, flag, sync, TRUE, buf);

	ahdb_unref (db);
	g_string_free (buf, TRUE);
}


/***
****
***/

static void
article_update_impl (article_db db,
		     const Article * article,
		     gboolean sync,
		     gboolean update_ui,
		     GString * buf)
{
	/* entry assertions */
	g_return_if_fail (article && article->message_id);

	/* update article header db... */
	ahdb_save_with_buffer (db, article, sync, buf);

	/* update the article ui... */
	if (update_ui)
		articlelist_update_node (article);
}

void
article_update (const Server *server,
		const Group *group,
		const Article* article,
		gboolean sync)
{
	GString * buf = g_string_new (NULL);
	article_db db = ahdb_ref (server, group);

	article_update_impl (db, article, sync, TRUE, buf);

	ahdb_unref (db);
	g_string_free (buf, TRUE);
}

/**
***
**/

gchar*
article_get_thread_message_id (const Article* article)
{
	char * pch;

	g_return_val_if_fail (article!=NULL, NULL);

	if (article->references)
	{
		pch = get_next_token_str (article->references, ' ', NULL);
	}
	else /* top of the thread */
	{
		pch = g_strdup (article->message_id);
	}

	return pch;
}

static int
article_get_entire_thread_count (Article* top)
{
	int i = 0;
	GSList *l;

	for (l=top->threads; l!=NULL; l=l->next)
		i += article_get_entire_thread_count (ARTICLE(l->data));

	return i;
}
static void
article_get_entire_thread_impl (Article* top, GPtrArray* setme)
{
	GSList *l;
	g_ptr_array_add (setme, top);
	for (l=top->threads; l!=NULL; l=l->next)
		article_get_entire_thread_impl (ARTICLE(l->data), setme);
}

static void
article_get_thread_impl (Article * article,
                         GPtrArray * setme,
                         gboolean children_only)
{
	Article* top = article;

	if (top != NULL)
	{
		int qty;

		if (!children_only)
			while (top->parent != NULL)
				top = top->parent;

	       	qty = article_get_entire_thread_count (top);
		pan_g_ptr_array_reserve (setme, setme->len + qty);
		article_get_entire_thread_impl (top, setme);
		pan_warn_if_fail (setme->len != qty);
	}
}

void
article_get_subthread (Article* article, GPtrArray* setme)
{
	article_get_thread_impl (article, setme, TRUE);
}

void
article_get_entire_thread (Article* article, GPtrArray* setme)
{
	article_get_thread_impl (article, setme, FALSE);
}

#if 0
/*
 * 1. open source db
 * 2. copy article from source db to folder db
 * 3. erase article from source db
 *
 */
static void
article_move_to_folder (Article *article,
			Server  *server,
			Group   *group,
			Group   *folder)
{
	g_return_if_fail (server != NULL);
	g_return_if_fail (article != NULL);
	g_return_if_fail (group != NULL);
	g_return_if_fail (folder != NULL);

	article_copy_to_folder (article, group, folder);

	article_delete (server, group, article, FALSE);
}

void
article_copy_to_folder (Article *article,
			Group   *group,
			Group   *folder)
{
//	article_db srcdb  = NULL;
	article_db destdb = NULL;

//	g_return_if_fail (server != NULL);
	g_return_if_fail (article != NULL);
	g_return_if_fail (group != NULL);
	g_return_if_fail (folder != NULL);

//	srcdb = ahdb_ref (server, from);
	destdb = ahdb_ref (folder_server, folder);

	ahdb_save (destdb, article, TRUE);

//	ahdb_unref (srcdb);
//	ahdb_sync (destdb);
	ahdb_unref (destdb);
}


typedef struct
{
	Server *server;
	Group  *group;
	GSList *list_of_articles;
} ArticleMoveStruct;

void
article_copy_list_to_folder (GSList *articles,
			     Group *group)
{

}

static void
article_move_list_destroy_cb (GtkObject *object, gpointer user_data)
{
	ArticleMoveStruct *ams = (ArticleMoveStruct*) user_data;
	GSList *l;

	for (l=ams->list_of_articles; l!=NULL; l=l->next)
	{
		GSList *l2 = (GSList*) l->data;
		g_slist_foreach (l2, article_free_gfunc, NULL);
		g_slist_free (l2);
	}
	g_slist_free (ams->list_of_articles);
	g_free (ams);
}

static void
article_move_list_clicked_cb (GnomeDialog *dialog,
                              int button,
                               gpointer user_data)
{
	ArticleMoveStruct *ams = (ArticleMoveStruct*) user_data;
	Server *server = ams->server;
	Group  *group = ams->group;
	Group *folder = NULL;
	GSList *l;

	g_return_if_fail (server != NULL);
	g_return_if_fail (group != NULL);


	/* Cancel button */
	if (button == 1)
	{
		pan_lock ();
		article_move_list_destroy_cb (GTK_OBJECT (dialog), NULL);
		gtk_widget_destroy (GTK_WIDGET (dialog));
		pan_unlock ();
		return;
	}

	folder = gtk_object_get_data (GTK_OBJECT (dialog), "selectedfolder");

	for (l=ams->list_of_articles; l!=NULL; l=l->next)
	{
		GSList *l2 = (GSList*) l->data;
		GSList *tmp = NULL;

		for (tmp=l2; tmp!=NULL; tmp=tmp->next)
		{
			Article *article = ARTICLE(tmp->data);

			article_move_to_folder (article, server, group, folder);
		}
	}

	pan_lock ();
	gtk_widget_destroy (GTK_WIDGET (dialog));
	pan_unlock ();
}

void
article_move_list_to_folder (GSList *articles,
                             Server *server,
                             Group *group)
{
	GnomeDialog *dialog = folderlist_picker_new ();

	ArticleMoveStruct *ams = g_new (ArticleMoveStruct, 1);

	ams->server = server;
	ams->group = group;
	ams->list_of_articles = articles;

	/* connect signals */
	pan_lock();
	gtk_signal_connect (GTK_OBJECT (dialog), "clicked",
			    GTK_SIGNAL_FUNC (article_move_list_clicked_cb),
			    ams);
//	gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
//			    GTK_SIGNAL_FUNC (article_move_list_destroy_cb),
//			    ams);
	pan_unlock();
			    
	gui_popup_draw (GTK_WIDGET (dialog), Pan.window);
}
#endif
