/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <string.h>
#include <stdlib.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnomeui/gnome-dialog.h>
#include <libgnomeui/gnome-stock.h>
/* #include <libgnomeui/libgnomeui.h> */

#include "articlelist.h"
#include "folderlist.h"
#include "globals.h"
#include "group.h"
#include "grouplist.h"
#include "prefs.h"
#include "gui.h"
#include "queue.h"
#include "server.h"
#include "task-bodies.h"
#include "task-headers.h"
#include "task-grouplist.h"
#include "util.h"


static void real_group_empty (Server *server, Group *group);

static GMemChunk * group_chunk = NULL;


Group*
group_new (void)
{
	Group *group;

	if (!group_chunk)
		group_chunk = g_mem_chunk_create (
			Group, 500, G_ALLOC_AND_FREE);

        group = g_chunk_new0 (Group, group_chunk);
	group->state_filter = ~STATE_FILTER_KILLFILE;
	group->type = GROUP;

	return group;
}

void
group_free (Group *group)
{
	g_return_if_fail (group!=NULL);

	g_free (group->name);
	g_free (group->description);
	g_chunk_free (group, group_chunk);
}

#if 0
Group*
group_copy (const Group *src)
{
	Group *dest = NULL;

	g_return_val_if_fail (src!=NULL, NULL);

	dest = group_new();
	dest->name = g_strdup (src->name);
	dest->description = g_strdup (src->description);
	dest->flags = src->flags;

	return dest;
}
#endif

gulong
group_get_attrib_ul (const Server   * server,
                     const Group    * group,
                     const char     * attribute)
{
	gulong retval = 0;
	char * pch;
       
	pch = group_get_attrib_string (server, group, attribute);
	if (pch != NULL)
	{
		retval = strtoul (pch, NULL, 10);
		g_free (pch);
	}

	return retval;
}

void
group_set_attrib_ul (const Server   * server,
                     const Group    * group,
                     const char     * attribute,
                     gulong           value)
{
	char * value_str = g_strdup_printf ("%lu", value);
	group_set_attrib_string (server, group, attribute, value_str);
	g_free (value_str);
}


void
group_set_attrib_string (const Server *server, const Group *group, const char *attrib, const char *value)
{
	if ( group->name )
	{
		gchar* key = g_strdup_printf ( "%s.%s", group->name, attrib );
		pan_db_put_value_str ( server->db, key, value );
		pan_db_sync ( server->db );
		g_free ( key );
	}
}

gchar*
group_get_attrib_string (const Server *server, const Group *group, const char *attrib)
{
	if ( !group->name )
		return NULL;
	else {
		gchar* key = g_strdup_printf ( "%s.%s", group->name, attrib );
		gchar* retval = pan_db_get_value_str ( server->db, key );
		g_free ( key );
		return retval;
	}
}

void
group_set_attrib_i (const Server *server, const Group *group, const char *attrib, int value)
{
	if ( group && group->name && attrib )
	{
		gchar* key = g_strdup_printf ( "%s.%s", group->name, attrib );
		pan_db_put_value_i ( server->db, key, value );
		pan_db_sync ( server->db );
		g_free ( key );

		/* FIXME this isn't broken, but it's kind of a kludge */
		if ( !strcmp(attrib,"Total") || !strcmp(attrib,"Read") )
			grouplist_update_row ( group );
	}
}

int
group_get_attrib_i (const Server *server, const Group *group, const char *attrib)
{

	if ( server && group && group->name && attrib )
	{
		gchar* key = g_strdup_printf ( "%s.%s", group->name, attrib );
		int retval = pan_db_get_value_i ( server->db, key );
		g_free ( key );
		return retval;
	}
	else return 0;
}

void
group_add_attrib_i (const Server *server, const Group *group, const char *attrib, int i)
{
	if ( server && group && group->name && attrib && i )
	{
		gchar* key = g_strdup_printf ( "%s.%s", group->name, attrib );
		pan_db_add_value_i ( server->db, key, i );
		pan_db_sync ( server->db );
		g_free ( key );

		/* Make the change visible */
		grouplist_update_row (group);
	}
}

int
compare_group (const Group *a, const Group *b)
{
	return strcmp(a->name, b->name);
}

int
compare_group_std (const void* va, const void* vb)
{
	return compare_group (*(const Group**)va, *(const Group**)vb);
}

static void
choose_download_dialog_clicked_cb (GnomeDialog *dialog,
				   int button_number,
				   gpointer group)
{
	Server* server =
		(Server*)gtk_object_get_data(GTK_OBJECT(dialog),"server");
	g_assert (server!=NULL);

	switch (button_number)
	{
		case 0: /* All Headers */
			queue_add(TASK(task_headers_new(
					server, group, HEADERS_ALL, TRUE)));
			break;

		case 1: /* New Headers */
			queue_add(TASK(task_headers_new(
					server, group, HEADERS_SAMPLE, TRUE)));
			break;

		default: /* Cancel */
			break;
	};

	gtk_widget_destroy (GTK_WIDGET(dialog));
}

void
choose_download_dialog (const Server *server,
			Group        *group)
{
	GtkWidget *dialog = NULL;
	GtkWidget *label = NULL;
	char *temp = NULL;

	pan_lock();

	dialog = gnome_dialog_new (_("Select Header Download Option"),
				   _("Download All Headers"),
				   _("Sample Some Headers"),
				   GNOME_STOCK_BUTTON_CANCEL, NULL);

	gtk_signal_connect (GTK_OBJECT(dialog), "clicked",
			    GTK_SIGNAL_FUNC (choose_download_dialog_clicked_cb),
			    group);
	gtk_object_set_data(GTK_OBJECT(dialog), "server", (gpointer)server);

	temp = g_strdup_printf (_("Select an option for group: %s"), group->name);
	label = gtk_label_new (temp);
	g_free (temp);

	gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(dialog)->vbox), label);
	pan_unlock();
	gui_popup_draw (dialog, Pan.window);
}

/* empty the group, delete the group database file and gnome_config sections */
static void
real_group_empty (Server *server,
		  Group  *group)
{
	if (group)
	{
		/* unload group (if it's loaded) */
		if (articlelist_get_current_group() == group)
			articlelist_clear ();

		/* clear out the article headers */
		ahdb_delete_db_file ( server, group );

		/* wipe out all records */
		group_set_attrib_i (server, group, "Total", 0);
		group_set_attrib_i (server, group, "Read", 0);

		/* flush the flags */
		if (group->flags & GROUP_SUBSCRIBED)
			group->flags = GROUP_SUBSCRIBED;
		else
			group->flags = 0;

		/* update display */
		grouplist_update_row (group);
	}
}

/* callback for button clicks in group_empty_dialog */
static void
group_empty_cb (GnomeDialog *dialog,
		int button_number,
		gpointer group)
{
	switch (button_number)
	{
		case 0: /* 'yes' */
			real_group_empty (grouplist_get_current_server (),
					  group);
			break;

		case 1: /* 'no' */
			break;

		default: /* cancel */
			break;
	};

	gtk_widget_destroy (GTK_WIDGET(dialog));
}

/**
 * group_empty_dialog:
 * @group: group to be emptied
 *
 * dialog popup asking user if they'd like to remove all the messages
 * for a given group.  called when the user tries to unsubscribe
 * or they choose the menu option to empty the group manually
 **/
void
group_empty_dialog (Group *group)
{
	GtkWidget *dialog = NULL;
	GtkWidget *label = NULL;
	gchar *pch = NULL;

	/* sanity check */
	g_return_if_fail (group!=NULL);

	pan_lock();

	/* create dialog */
	pch = g_strdup_printf (_("Clean group ``%s''"), group->name);
	dialog = gnome_dialog_new (pch,
				   GNOME_STOCK_BUTTON_YES,
				   GNOME_STOCK_BUTTON_NO, NULL);
	g_free (pch);

	/* body */
	pch = g_strdup_printf (_("Remove all messages from group: \n%s ?"), group->name);
	label = gtk_label_new (pch);
	g_free (pch);

	/* listen for clicks */
	gtk_signal_connect (GTK_OBJECT(dialog), "clicked",
			    GTK_SIGNAL_FUNC (group_empty_cb), group);

	gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(dialog)->vbox), label);
	pan_unlock();

	gui_popup_draw (dialog, Pan.window);
}

/*
 * 1. remove article db file
 * 2. remove group from the {group/folder}list
 * 3. save list
 */
static void
real_group_remove (Server *server, Group *group)
{
	/* clear out the article headers */
	ahdb_delete_db_file (server, group);
	g_message ("real_group_remove, type = %d", group->type);
	
	grouplist_remove_row (group);

	if (group->type == USER_FOLDER)
	{
		pan_db db;
		gchar *path;

		/* remove folder from folderlist */
		path = g_strdup_printf ("/%s/folderlist.db", data_dir);
		db = pan_db_open (path);
		pan_db_erase (db, group->name);
		pan_db_close (db);
		db = NULL;
		g_message (path);
		g_free (path);

		/* remove the folder file */
		path = g_strdup_printf ("/%s/folders/%s.db", data_dir, group->name);
		g_message ("db filename: %s", path);
		remove (path);
		g_free (path);

		/* remove from current Pan session */
		g_ptr_array_remove (folders, group);
	}
	else if (group->type == GROUP)
	{
		pan_db db;
		gchar *path = g_strdup_printf ("/%s/%s_grouplist.db", data_dir, server->name);
		db = pan_db_open (path);
		g_free (path);

		g_ptr_array_remove (server->groups, group);
		pan_db_erase (db, group->name);
		pan_db_close (db);
	}
}

/* callback for button clicks in group_remove_dialog */
static void
group_remove_cb (GnomeDialog *dialog,
		 int button_number,
		 gpointer group)
{
	switch (button_number)
	{
		case 0: /* 'yes' */
			real_group_remove (grouplist_get_current_server (),
					   group);
			break;

		case 1: /* 'no' */
			break;

		default: /* cancel */
			break;
	};

	gtk_widget_destroy (GTK_WIDGET(dialog));
}


void
group_remove_dialog (Group *group)
{
	GtkWidget *dialog = NULL;
	GtkWidget *label = NULL;
	gchar *pch = NULL;

	/* sanity check */
	g_return_if_fail (group!=NULL);
	g_return_if_fail (group->name != NULL);

	/* create dialog */
	pch = g_strdup_printf (_("Remove group ``%s''"), group->name);
	dialog = gnome_dialog_new (pch,
				   GNOME_STOCK_BUTTON_YES,
				   GNOME_STOCK_BUTTON_NO, NULL);
	g_free (pch);

	/* body */
	pch = g_strdup_printf (_("Remove group: \n%s ?"), group->name);
	label = gtk_label_new (pch);
	g_free (pch);

	/* listen for clicks */
	gtk_signal_connect (GTK_OBJECT(dialog), "clicked",
			    GTK_SIGNAL_FUNC (group_remove_cb), group);

	gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(dialog)->vbox), label);

	gui_popup_draw (dialog, Pan.window);
}

void
group_download_all_articles (Server   * server,
                             Group    * group)
{
	article_db db = ahdb_ref (server, group);
	if (db != NULL)
	{
		GPtrArray * articles = ahdb_get_all (db, NULL, NULL);
		PanObject * o = task_bodies_new (server, group, articles);
		queue_add (TASK(o));
		ahdb_unref (db);
	}
}

