/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "acache.h"
#include "article-db.h"
#include "date.h"
#include "globals.h"
#include "log.h"
#include "nntp.h"
#include "queue.h"

#include "util.h"

gboolean online_prompt = TRUE;

extern const char* sockwrite_err_msg;
extern const char* sockread_err_msg;

void
nntp_disconnect (PanSocket *sock)
{
	pan_socket_putline (sock, "QUIT\r\n");
}

const gchar*
nntp_login (PanSocket *sock,
            Server* server)
{
	int ival;
	const gchar* buffer;

	g_return_val_if_fail (sock!=NULL, NULL);

	if (pan_socket_getline (sock, &buffer))
		return _("Error reading from socket.");

	/* attempt authentication.  This isn't implemented exactly as
	 * "Common NNTP Extensions" specifies, being that we didn't
	 * wait to be asked for it, but hopefully it will work.
	 */
	if (server->need_auth)
	{
		if (pan_socket_putline_va (sock, "AUTHINFO user %s\r\n", server->username))
			return _("Error writing to socket.");
		if (pan_socket_getline (sock, &buffer))
			return _("Error reading from socket.");
		if (atoi(buffer) == 381)
		{
			/* 381 Means: More authentication information needed */
			if (pan_socket_putline_va (sock, "AUTHINFO pass %s\r\n", server->password))
				return _("Error writing to socket.");

			if (pan_socket_getline (sock, &buffer))
				return _("Error reading from socket.");

			/* 281 Means: Authentication accepted */
			if (atoi(buffer) != 281) {
				log_add_va ( _("Authentication for server '%s' unsuccessful."),
					server->name);
				return buffer;
			}
		}
		/* Else authentication not required or not understood */
	}

	if (pan_socket_putline (sock, "MODE READER\r\n"))
		return _("Error writing to socket.");
	if (pan_socket_getline (sock, &buffer))
		return _("Error reading from socket.");

	ival = atoi(buffer);
	if (ival == 200)
		server->posting = TRUE;
	else if (ival == 201)
		server->posting = FALSE;
	else if (ival == 500) /* command not recognized; "mode reader" is an extension */
		server->posting = TRUE; /* a wild guess */
	else
		return buffer;

	return NULL;
}




int
nntp_set_group (PanSocket    * sock,
                const char   * group_name)
{
	const gchar *buffer = NULL;

	g_return_val_if_fail (is_nonempty_string(group_name), -1);

	/* write the request */
	if (pan_socket_putline_va (sock, "GROUP %s\r\n", group_name))
		return -1;

	/* get the result string */
	if (pan_socket_getline (sock, &buffer))
		return -1;

	/* interpret the result string */
	if (atoi(buffer)!=210 && atoi(buffer)!=211) {
		g_warning (_("Server returned `%s'"), buffer);
		pan_error_dialog (_("Server returned `%s'"), buffer);
		return -1;
	}

	return 0;
}



/**
 * nntp_post:
 * @msg: ???
 *
 * Return value:
 **/
int
nntp_post (PanSocket *sock,
           const gchar *msg)
{
	const gchar *buffer;

	if (Pan.mute) {
		g_message ("Mute: Your Mesage will not actually be posted.");
		printf ("\n\n\nYour message\n%s<end of message>\n\n\n", msg);
		fflush (NULL);
		return 0;
	}

	if (pan_socket_putline (sock, "POST\r\n")) {
		pan_error_dialog (_("Error writing to socket during post."));
		return -1;
	}

	if (pan_socket_getline (sock, &buffer)) {
		pan_error_dialog (_("Error reading from socket during post."));
		return -1;
	}

	if (atoi(buffer) == 340) { /* 340 Means: Send Article To Be Posted */
		if (pan_socket_putline_va (sock, "%s\r\n.\r\n", msg)) {
			pan_error_dialog (_("Error writing to socket during post."));
			return -1;
		}
		if (pan_socket_getline (sock, &buffer)) {
			pan_error_dialog (_("Error reading from socket during post."));
			return -1;
		}
		if (atoi(buffer) == 240) { /* 240 Means: Article Posted OK */
			return 0;
		}
	}

	pan_error_dialog (_("Error posting message.  Server said:\n'%s'"), buffer);
	return -1;
}

/*--------------------------------------------------------------------
 * download and decode an article
 *--------------------------------------------------------------------*/

gint
nntp_send_noop (PanSocket *sock)
{
	const char *buffer;

	/* send the request.  Using "mode reader"
	   because it's a lightweight request */
	if (pan_socket_putline (sock, "MODE READER\r\n")) {
		pan_error_dialog (_("Error writing to socket during no-op."));
		return -1;
	}

	/* get the response */
	if (pan_socket_getline (sock, &buffer)) {
		pan_error_dialog (_("Error reading from socket during no-op."));
		return -1;
	}

	/* we don't care what the response is, we're just doing this to keep
	   the connection alive.  So don't parse it. */
	return 0;
}


int
nntp_article_download (StatusItem* item,
                       PanSocket* sock,
                       const Server* server,
                       const Group* group,
                       Article* article,
                       const gboolean* abort,
                       gboolean verbose)
{
	gboolean article_headers_changed = FALSE;
	const char *buffer;
	FILE *f;

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, -1);

	if (verbose)
		status_item_emit_status (
			STATUS_ITEM(item),
			_("Downloading article '%s'"),
			article->subject);

	/* get the head */
	if (*abort || pan_socket_putline_va (sock, "HEAD %lu\r\n", article->number))
	{
		status_item_emit_error (STATUS_ITEM(item), sockwrite_err_msg);
		return -1;
	}
	if (*abort || pan_socket_getline (sock, &buffer))
	{
		status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
		return -1;
	}
	if (atoi(buffer) != 221)
	{
		char* pch = g_strdup (buffer);
		pch[strcspn(pch,"\r\n")] = '\0';
		status_item_emit_error(STATUS_ITEM(item), _("NNTP \"HEAD\" command returned `%s'."), pch);
		g_free (pch);
		return -2;
	}
	if (*abort || pan_socket_getline (sock,&buffer))
	{
		status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
		return -1;
	}

	f = acache_open_header (article);
	while (strncmp(buffer, ".\r\n", 3))
	{
		char* pch = NULL;
		const char* name = NULL;

		/* check for exit... */
		if (*abort) {
			acache_close_header (article, f);
			acache_delete (article->message_id);
			return -1;
		}

		/* write to acache... */
		acache_putline_header (f, buffer);

		/* parse it for article... */
		if ((pch = strrchr (buffer, '\x0D'))) /* trim linefeed */
			*pch = '\0';
		name = "Newsgroups: ";
		if (!strncmp ( buffer, name, strlen(name))) {
			g_free (article->newsgroups);
			article->newsgroups = g_strdup ( buffer+strlen(name) );
			article_headers_changed = TRUE;
		}
		name = "Reply-To: ";
		if (!strncmp (buffer, name, strlen(name))) {
			g_free (article->reply_to);
			article->reply_to = g_strdup ( buffer+strlen(name) );
			article_headers_changed = TRUE;
		}
		name = "Followup-To: ";
		if (!strncmp ( buffer, name, strlen(name))) {
			g_free (article->followup_to);
			article->followup_to = g_strdup ( buffer+strlen(name) );
			article_headers_changed = TRUE;
		}

		/* get the next line... */
		if (*abort || pan_socket_getline (sock, &buffer)) {
			status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
			acache_close_header (article, f);
			acache_delete (article->message_id);
			return -1;
		}
	}
	acache_close_header (article, f);

	if (article_headers_changed)
		article_update (server, group, article, TRUE);

	if (*abort || pan_socket_putline_va (sock, "BODY %lu\r\n", article->number)) {
		status_item_emit_error (STATUS_ITEM(item), sockwrite_err_msg);
		return -1;
	}
	if (*abort || pan_socket_getline (sock, &buffer)) {
		status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
		return -1;
	}
	if (atoi(buffer) != 222) {
		return -2;
	}
	
	if (*abort || pan_socket_getline (sock, &buffer)) {
		status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
		return -1;
	}
	f = acache_open_body (article);

	if (verbose)
		status_item_emit_init_steps (STATUS_ITEM(item), article->linecount);

	while (strncmp(buffer, ".\r\n", 3))
	{
		if (*abort)
		{
			acache_close_body (article, f);
			acache_delete (article->message_id);
			return -1;
		}
		acache_putline_body (f, buffer);

		if (verbose)
			status_item_emit_next_step(STATUS_ITEM(item));

		if (*abort || pan_socket_getline (sock, &buffer))
		{
			status_item_emit_error (STATUS_ITEM(item), sockread_err_msg);
			acache_close_body (article, f);
			acache_delete (article->message_id);
			return -1;
		}
	}
	acache_close_body (article, f);
	acache_file_checkin (article->message_id);

	if (verbose)
		status_item_emit_progress(STATUS_ITEM(item), 100);

	return 0;
}


int
nntp_get_group_info (StatusItem       * status,
                     PanSocket        * sock,
                     const char       * group_name,
                     unsigned long    * article_qty,
                     unsigned long    * low_num,
                     unsigned long    * high_num,
                     const gboolean   * abort)
{
	const char *buffer = NULL;

	/* change to the right group */
	if (*abort || pan_socket_putline_va (sock, "GROUP %s\r\n", group_name))
	{
		return -1;
	}
	if (*abort || pan_socket_getline (sock, &buffer))
	{
		status_item_emit_error (status, sockread_err_msg);
		return -1;
	}
	if (strtol(buffer, NULL, 10) != 211)
	{
		status_item_emit_error (status, _("(Couldn't select group '%s'"), group_name);
		return -1;
	}
	sscanf (buffer, "%*d %lu %lu %lu", article_qty, low_num, high_num);
	return 0;
}


int
nntp_articlelist_download (StatusItem       * item,
                           PanSocket        * sock,
                           const Server     * server,
                           const Group      * group,
                           unsigned long      low,
                           unsigned long      high,
                           const gboolean   * abort,
                           gboolean           verbose,
                           const char       * progress_fmt,
                           GPtrArray        * setme)
{
	Article *article;
	const char *buffer;
	unsigned long total;
	unsigned long first;
	unsigned long last;
	gboolean read_status = 0;

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, -1);
	g_return_val_if_fail (server!=NULL, -1);
	g_return_val_if_fail (group!=NULL, -1);

	/* change to the right group */
	total = first = last = 0;
	if (nntp_get_group_info (item, sock, group->name, &total, &first, &last, abort))
		return -1;

	/* if no articles to get... */
	if (total==0) {
		if (verbose)
			status_item_emit_status (item, _("No articles found for group '%s'"), group->name);
		return 0;
	}

	/* get the articles... */
	if (verbose)
		status_item_emit_status (item, _("Getting %d article headers..."), high-low);
	if (*abort || pan_socket_putline_va (sock, "XOVER %lu-%lu\r\n", low, high)) {
		task_emit_sockwrite_err(TASK(item));
		return -1;
	}

	/* wait for a reply... */
	if (verbose)
		status_item_emit_status (STATUS_ITEM(item), _("Waiting for Response from Server..."));
	if (*abort || pan_socket_getline (sock, &buffer)) {
		task_emit_sockread_err (TASK(item));
		return -1;
	}

	/* parse the reply... */
	if (strtol(buffer, (char **) NULL, 10) != 224) {
		status_item_emit_error (STATUS_ITEM(item),
			_("Couldn't retrieve headers\nServer said: %s"), buffer);
		return -1;
	}

	/* reserve len article pointers in the ptr array */
	//pan_g_ptr_array_reserve (setme, high-low);

	/* load the article headers */
	while (!*abort && !((read_status = pan_socket_getline (sock, &buffer))))
	{
		int i = 0;
		int part = 0;
	        int parts = 0;
		const char * pch;
		char *s = NULL;

		/* handle end-of-list */
		if (!strncmp(buffer, ".\r\n", 3))
			break;

		/* create the article data */
		pch = buffer;
		article = article_new ();
		article->number         = get_next_token_ulong (pch, '\t', &pch);
		article->subject        = get_next_token_str   (pch, '\t', &pch);
		article->author         = get_next_token_str   (pch, '\t', &pch);
		article->date           = parse_date (s=get_next_token_str(pch, '\t', &pch));
		article->message_id     = get_next_token_str   (pch, '\t', &pch);
		article->references     = get_next_token_str   (pch, '\t', &pch);
		                          get_next_token_int   (pch, '\t', &pch); /* unused field */
		article->linecount      = get_next_token_int   (pch, '\t', &pch);
		article->state          = STATE_NEW;
		article->crosspost_qty  = 1; /*FIXME -- Xref?*/
		g_free (s);
		s = NULL;

		/* s'more validity checking */
		if ((article->number==0)
			|| !is_nonempty_string (article->subject)
			|| !is_nonempty_string (article->author)
			|| !is_nonempty_string (article->message_id))
		{
			g_message (_("Corrupt Header"));
			status_item_emit_error (
				STATUS_ITEM(item),
				_("Corrupt header ``%s'' skipped"),
				buffer);
			pan_object_unref (PAN_OBJECT(article));
			continue;
		}

		/* Look for the (n/N) or [n/N] construct in subject lines,
		 * starting at the end of the string and working backwards */
		part = 0;
		parts = 0;
		s = article->subject;
		for (i = strlen(s) - 1; i > 0; --i) {
			if ((s[i] == ')') || (s[i] == ']')) {
				for (--i; isdigit((int)s[i]); --i);
				if (s[i] == '/') {
					parts = atoi(s + i + 1);
					for (--i; isdigit((int)s[i]); --i);
					if ((s[i] == '(') || (s[i] == '[')) {
						part = atoi(s + i + 1);
						break;
					}
				}
			}
		}

		/* if not a multipart yet, AND if it's a big message, AND
		   it's either in one of the pictures/fan/sex groups or it
		   has commonly-used image names in the subject, guess it's
		   a single-part binary */
		if (!parts
			&& article->linecount>400
			&& (((strstr(group->name,"pictures")
					|| strstr(group->name,"fan")
				        || strstr(group->name,"mag")
					|| strstr(group->name,"sex")))
				|| ((strstr(s,".jpg") || strstr(s,".JPG")
					|| strstr(s,".jpeg") || strstr(s,".JPEG")
					|| strstr(s,".gif") || strstr(s,".GIF")
					|| strstr(s,".tiff") || strstr(s,".TIFF")))))
			part = parts = 1;

		/* Verify Multipart info */
		if ((parts>=1) && (part<=parts)) {
			/* We found one! */
			article->parts = parts;
			article->part = part;
		}
		else {
			article->parts = 0;
			article->part = 0;
		}

		/* Add the article to the article list */
		g_ptr_array_add (setme, article);

		/* update the count & progress ui */
		if (verbose) {
			status_item_emit_next_step (STATUS_ITEM(item));
			if (!(setme->len % 29)) {
				status_item_emit_status (STATUS_ITEM(item),
					progress_fmt,
					setme->len, (high-low));
			}
		}
	}
	if (read_status!=0 || TASK(item)->hint_abort)
	{
		if (read_status!=0)
			task_emit_sockread_err (TASK(item));
		return -1;
	}

	return 0;
}


int
nntp_download_bodies (StatusItem * item,
                      Server * server,
                      Group * group,
                      PanSocket * sock,
                      const gboolean * abort,
                      const GPtrArray * const articles,
                      int * index)
{
	article_db db;

	/* sanity checks */
	g_return_val_if_fail (server!=NULL, -1);
	g_return_val_if_fail (group!=NULL, -1);
	g_return_val_if_fail (is_nonempty_string(group->name), -1);
	g_return_val_if_fail (articles!=NULL, -1);

	/* go to the right group */
	if (*abort || nntp_set_group (sock, group->name))
	{
		status_item_emit_error (item,
		                        _("Couldn't select group '%s'"),
		                        group->name);
		return -1;
	}


	/* download the articles */
	for (; *index<articles->len && !*abort; ++(*index))
	{
		Article* article = ARTICLE(g_ptr_array_index(articles,*index));
		int retval = 0;

		if (!acache_body_exists(article->message_id) ||
		    !acache_header_exists(article->message_id))
		{
			retval = nntp_article_download (item,
			                                sock,
			                                server,
			                                group,
			                                article,
			                                abort,
			                                FALSE);
		}

		if (retval)
		{
			status_item_emit_status (item,
			                         _("Downloading article ``%s'' failed."),
			                         article->subject);
		}
		else
		{
			status_item_emit_next_step (item);

			status_item_emit_status (item,
			                         _("Getting article %d of %d"),
			                         1+*index,
			                         articles->len);

			article_remove_flag (server,
			                     group,
			                     article,
			                     STATE_DOWNLOAD_QUEUED,
			                     FALSE);
		}
	}

	/* sync the article database */
	db = ahdb_ref (server, group);
	ahdb_sync (db);
	ahdb_unref (db);

	return 0;
}
