/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <string.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-config.h>
#include <libgnome/gnome-i18n.h>
#include <libgnomeui/libgnomeui.h>

#include "debug.h"
#include "globals.h"
#include "grouplist.h"
#include "gui.h"
#include "prefs.h"
#include "queue.h"
#include "server.h"
#include "util.h"

static void server_list_save (void);

/**
***  Life Cycle
**/

void
server_constructor (Server              * server,
                    PanObjectDestructor   destructor)
{
	PanObject *obj = PAN_OBJECT(server);

	/* init the superclass bits */
	pan_object_constructor (obj, destructor);

	/* init the status-item bits */
	debug (DEBUG_PAN_OBJECT, "server_constructor: %p", server);
	server->online_status_changed = pan_callback_new ();
	server->posting = FALSE;
	server->gen_msgid = FALSE;
	server->need_auth = FALSE;
	server->online = FALSE;
	server->online_prompted = FALSE;
	server->reserve_nonleech = TRUE;
	server->address = NULL;
	server->name = NULL;
	server->username = NULL;
	server->password = NULL;
	server->port = 25;
	server->max_connections = 1;
	server->idle_secs_before_timeout = 30;
	server->groups = NULL;
	server->db = NULL;
	server->last_newgroup_list_time = 1; /* Typhoon can't handle 0 */
}

void
server_destructor (PanObject *obj)
{
	Server* server = SERVER(obj);

	/* clear out the server bits */
	debug (DEBUG_PAN_OBJECT, "server_destructor: %p", obj);
	g_free (server->name);
	g_free (server->address);
	g_free (server->username);
	g_free (server->password);
	g_ptr_array_free (server->groups, TRUE);
	pan_callback_free (server->online_status_changed);
	if (server->db)
		pan_db_close (server->db);

	/* clear out the parent */
	pan_object_destructor (obj);
}

Server*
server_new (void)
{
	Server* server = g_new0 (Server, 1);
	debug (DEBUG_PAN_OBJECT, "server_new: %p", server);
	server_constructor (server, server_destructor);
	return server;
}

/**
***
**/

void
server_set_online (Server* server,
                   gboolean online)
{
	g_return_if_fail (server!=NULL);
	server->online = online ? 1 : 0;
	if (online)
		queue_wakeup ();
}

/**
***
**/

void
server_shutdown (void)
{
	GSList *l;
	Server *last;

	/* remember which server we're on so that
	   we can default to it next time... */
	last = grouplist_get_current_server ();
	if (last && is_nonempty_string(last->name))
	{
		gnome_config_set_string ("/Pan/State/Server", last->name);
		gnome_config_sync ();
	}

	/* save the updated server info */
	server_list_save ();

	/* close down the servers */
	l = Pan.serverlist;
	Pan.serverlist = NULL;
	g_slist_foreach (l, (GFunc)pan_object_unref, NULL);
	g_slist_free (l);
}



/**
***  GUI Stuff
**/

static void
server_selected_cb (GtkWidget  * w,
                    Server     * server)
{
	if (GTK_CHECK_MENU_ITEM(w)->active)
		grouplist_server_selected (server);
}

static GnomeUIInfo end = GNOMEUIINFO_END;


static int server_qty = 0;
static gchar * first_server_name = NULL;
static GnomeUIInfo * server_menu = NULL;
static GnomeUIInfo * server_menu_radiobox = NULL;

void
server_menu_update (void)
{
	GSList* l;
	GnomeUIInfo * buf;
	GnomeUIInfo * radio_box;
	GString * str = g_string_new (NULL);
	int i;
	gchar *path;

	/* build the new menu */
	l = Pan.serverlist;
	buf = g_new0 (GnomeUIInfo, g_slist_length(l)+2);
	i = 0;
	if (!l)
	{
		GnomeUIInfo * info = &buf[i];
		info->type = GNOME_APP_UI_ITEM;
		info->label = _("No Servers Defined");
		info->hint = _("No Servers Defined");
		info->moreinfo = NULL;
		info->user_data = NULL;
		++i;
	}
	for (; l!=NULL; l=l->next, ++i)
	{
		Server *s = SERVER(l->data);
		GnomeUIInfo * info = &buf[i];

		info->type = GNOME_APP_UI_ITEM;
		info->label = s->name;
		info->hint = s->name;
		info->moreinfo = server_selected_cb;
		info->user_data = s;

		/* open the server's db... */
		if (s->db == NULL)
		{
			g_string_sprintf(str,"/%s/%s.db",data_dir,s->name);
			s->db = pan_db_open (str->str);
		}
	}
	buf[i] = end;
	radio_box = g_new0 (GnomeUIInfo, 2);
	radio_box->type = GNOME_APP_UI_RADIOITEMS;
	radio_box->label = _("Servers");
	radio_box->hint = _("Servers");
	radio_box->moreinfo = buf;
	radio_box[1] = end;


	/* update the main menu */
	if (first_server_name)
	{
		gchar* pch = g_strdup_printf ("Server/%s", first_server_name);
		gnome_app_remove_menus (GNOME_APP(Pan.window), pch, server_qty);
		g_free (pch);
	}

	path = g_strdup_printf ("%s/%s", _("_Server"), "<Separator>");
	gnome_app_insert_menus (GNOME_APP(Pan.window), path, radio_box);
	g_free (path);

	/* update the local varialbes */
	g_free (first_server_name);
	g_free (server_menu);
	g_free (server_menu_radiobox);
	first_server_name = g_strdup (Pan.serverlist ? SERVER(Pan.serverlist->data)->name : NULL);
	server_menu = buf;
	server_menu_radiobox = radio_box;
	server_qty = i;

	/* save groups to disk */
	server_list_save ();

	/* cleanup */
	g_string_free (str, TRUE);
}

Server*
server_fetch (const char *name)
{
	Server *server = NULL;
	gboolean has_section;

	gchar* buf = g_strdup_printf ( "/Pan/%s/", name );
	gnome_config_push_prefix ( buf );
	has_section = gnome_config_has_section ( buf );
	g_free ( buf );

	if ( has_section )
	{
		server = server_new ();
		server->address = gnome_config_get_string ("Address");
		if (server->address == NULL)
		{
			g_free (server);
			server = NULL;
		}
		else
		{
			gchar *path;
			gchar *pch;
			
			/* make sure the server's config directory exists */
		       	path = g_strdup_printf ("/%s/%s/", data_dir, name);
			directory_check (path);
			g_free (path);

			/* load the server from gnome_config */
			server->port =
				gnome_config_get_int ("Port=119");
			server->name =
				g_strdup(name);
			server->gen_msgid =
				gnome_config_get_int ("Generate_Msgid=0");
			server->need_auth =
				gnome_config_get_int ("Auth_Required=0");
			server->username =
				gnome_config_get_string ("Username");
			server->password =
				gnome_config_get_string ("Password");
			server->max_connections =
				gnome_config_get_int ("Max_Connections=2");
			server->idle_secs_before_timeout =
				gnome_config_get_int ("Idle_Timeout_Secs=180");
			server->online =
				gnome_config_get_bool ("Online=TRUE");
			server->reserve_nonleech =
				gnome_config_get_bool ("Reserve_Nonleech_Connection=TRUE");
			/* no good way to get a ulong from gnome_config? */
			server->last_newgroup_list_time = 1;
			pch = gnome_config_get_string ("Last_Grouplist_Update");
			if (is_nonempty_string(pch))
				server->last_newgroup_list_time = strtoul (pch, NULL, 10);
			g_free (pch);

			server->groups = g_ptr_array_new ();
		}
	}
	gnome_config_pop_prefix ();

	return server;
}


void
server_init (void)
{
	gchar *last = NULL;
	GSList *l = NULL;
	int i;

	/* This gets called before gtk_main and
	 * at the end of guru_finish */
	Pan.serverlist = server_list_load ();
	server_menu_update ();

	/* try hard to select a server.. if there's one in the ini file, use
	 * it; otherwise, go with the first one in the list (if there is one)
	 */
	last = gnome_config_get_string ("/Pan/State/Server");
	if (!last && Pan.serverlist)
		last = g_strdup ((SERVER(Pan.serverlist->data)->name));
	if (last!=NULL)
	{
		for (i=0, l=Pan.serverlist; l; l=l->next, ++i)
		{
			Server *server = SERVER(l->data);
			if (!strcmp(server->name, last))
			{
				gtk_check_menu_item_set_active (
		                	GTK_CHECK_MENU_ITEM(
						server_menu[i].widget), TRUE);
				grouplist_server_selected (server);
				break;
			}
		}
		g_free (last);
	}
}


static void
server_list_save (void)
{
	GString * buf = g_string_new (NULL);
	GString * servers = g_string_new (NULL);
	GSList *l = NULL;
	gchar * p;

	/* Out with the old */
	p = gnome_config_get_string ("/Pan/Servers/Names");
	if (p != NULL) {
		const char * pch = p;
		while (get_next_token_g_str(pch, ',', &pch, buf)) {
			g_string_prepend (buf, "/Pan/");
			gnome_config_clean_section (buf->str);
		}
		g_free (p);
	}
	gnome_config_sync ();

	/* In with the new */
	for (l=Pan.serverlist; l!=NULL; l=l->next)
	{
		char * pch;
		Server *s = SERVER(l->data);

		g_string_sprintf (buf, "/Pan/%s/Address", s->name);
		gnome_config_set_string (buf->str, s->address);

		g_string_sprintf (buf, "/Pan/%s/Port", s->name );
		gnome_config_set_int (buf->str, s->port);

		g_string_sprintf (buf, "/Pan/%s/Auth_Required", s->name);
		gnome_config_set_int (buf->str, s->need_auth);

		g_string_sprintf (buf, "/Pan/%s/Generate_Msgid", s->name);
		gnome_config_set_int (buf->str, s->gen_msgid);

		g_string_sprintf (buf, "/Pan/%s/Username", s->name);
		gnome_config_set_string (buf->str, s->username);

		g_string_sprintf (buf, "/Pan/%s/Password", s->name);
		gnome_config_set_string (buf->str, s->password);

		g_string_sprintf (buf, "/Pan/%s/Max_Connections", s->name);
		gnome_config_set_int (buf->str, s->max_connections);

		g_string_sprintf (buf, "/Pan/%s/Idle_Timeout_Secs", s->name);
		gnome_config_set_int (buf->str, s->idle_secs_before_timeout);

		g_string_sprintf (buf, "/Pan/%s/Online", s->name);
		gnome_config_set_bool (buf->str, s->online);

		/* no good way to store a ulong in gnome_config? */
		pch = g_strdup_printf ("%lu", s->last_newgroup_list_time);
		g_string_sprintf (buf,"/Pan/%s/Last_Grouplist_Update", s->name);
		gnome_config_set_string (buf->str, pch);
		g_free (pch);

		g_string_sprintf (buf, "/Pan/%s/Reserve_Nonleech_Connection", s->name);
		gnome_config_set_bool (buf->str, s->reserve_nonleech);

		if (servers->len)
			g_string_append_c (servers, ',');
		g_string_append (servers, s->name);
	}

	gnome_config_set_string ("/Pan/Servers/Names", servers->str);
	g_string_free (servers, TRUE);
	g_string_free (buf, TRUE);

	gnome_config_sync ();
}

static gint
server_name_compare (gconstpointer a,
                     gconstpointer b)
{
	const Server* sa = (const Server*) a;
	const Server* sb = (const Server*) b;
	return g_strcasecmp (sa->name, sb->name);
}

/**
 * server_list_load:
 * Load a list of servers from the user's ~/.gnome/Pan file
 * @return a GSList of Server's for each server found
 **/
GSList *
server_list_load (void)
{
	gchar *p;
	GSList *l = NULL;

	/* get a comma-delimited list of all the server names */
	p = gnome_config_get_string ("/Pan/Servers/Names");
	if (p != NULL)
	{
		const char * pch;
		GString * str = g_string_new (NULL);

		/* create a server object for each server in the list */
		pch = p;
		while (get_next_token_g_str (pch, ',', &pch, str))
		{
			Server * server;
			server = server_fetch (str->str);
			if (server)
				l = g_slist_insert_sorted (l, server, server_name_compare);
		}

		/* cleanup */
		g_string_free (str, TRUE);
		g_free (p);
	}

	return l;
}




static int
compare_name_to_group (const void * a,
                       const void * b)
{
	return strcmp ((const char*)a, (*(Group**)b)->name);
}

gboolean
server_has_group (const Server  * server,
                  const char    * group_name)
{
	gboolean exact_match = FALSE;

	g_return_val_if_fail (server!=NULL, FALSE);
	g_return_val_if_fail (group_name!=NULL, FALSE);
	g_return_val_if_fail (*group_name!='\0', FALSE);

	lower_bound (group_name,
                     server->groups->pdata,
                     server->groups->len,
                     sizeof(gpointer),
                     compare_name_to_group,
                     &exact_match);

	return exact_match;
}

/**
***
**/

static void
online_prompt_cb (gint reply,
                  gpointer user_data)
{
	const gboolean online = !reply;
	Server * server = SERVER(user_data);
	if (server->online != online)
	{
		server_set_online (SERVER(user_data), online);
		server->online_prompted = FALSE;
	}
}

void
server_prompt_to_go_online (Server* server)
{
	g_return_if_fail (server!=NULL);

	if (!server->online && !server->online_prompted)
	{
		gchar* question;

		server->online_prompted = TRUE;

		question = g_strdup_printf (_("Go Online for Server `%s'?"),
		                            server->name);

		pan_lock ();
		gnome_question_dialog_parented (question,
		                                online_prompt_cb,
		                                server,
		                                GTK_WINDOW(Pan.window));
		pan_unlock ();

		g_free (question);
	}
}
