/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "article.h"
#include "article-db.h"
#include "date.h"
#include "debug.h"
#include "log.h"
#include "nntp.h"
#include "task-headers.h"

#include "articlelist.h"
#include "grouplist.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static gint task_headers_run (Task* item);

static gchar* task_headers_describe (const StatusItem *item);

static int task_get_header_range (TaskHeaders  * task,
                                  unsigned long * setme_low,
                                  unsigned long * setme_high,
                                  unsigned long * setme_first,
                                  unsigned long * setme_last,
		                  unsigned long * setme_qty,
		                  const char ** setme_progress_fmt);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

/*****
******
*****/

static void
task_headers_destructor (PanObject* o)
{
	TaskHeaders * task_h = TASK_HEADERS(o);

	/* destruct this class... */
	pan_g_ptr_array_foreach (task_h->articles, (GFunc)pan_object_unref, 0);
	g_ptr_array_free (task_h->articles, TRUE);

	/* destruct parent class... */
	task_destructor (o);
}

static PanObject*
real_task_headers_new (Server               * server,
                       Group                * group,
		       HeaderDownloadType     dl_type,
                       gboolean               high_priority,
		       gboolean               download_bodies)
{
	TaskHeaders *item = g_new0 (TaskHeaders, 1);
	gboolean need_socket = dl_type != HEADERS_LOCAL;

	debug (DEBUG_PAN_OBJECT, "task_headers_new: %p", item);

	task_constructor (TASK(item),
		task_headers_destructor,
		task_headers_describe,
		task_headers_run, server, high_priority, need_socket);

	item->group = group;
	item->download_type = dl_type;
	item->download_bodies = download_bodies;
	item->articles = g_ptr_array_new ();
	item->body_index = 0;

	return PAN_OBJECT(item);
}

PanObject* task_headers_new (Server              * server,
                             Group               * group,
			     HeaderDownloadType    dl_type,
			     gboolean              high_priority)
{
	return real_task_headers_new (server,
	                              group,
	                              dl_type,
	                              high_priority,
	                              FALSE);
}

PanObject* task_headers_new_with_bodies   (Server              * server,
                                           Group               * group,
					   HeaderDownloadType    dl_type)
{
	return real_task_headers_new (server,
	                              group,
	                              dl_type,
	                              FALSE,
	                              TRUE);
}

/*****
******
*****/

static int
pp_article_message_id_comp (const void* va,
                            const void* vb )
{
	const Article* a = *(const Article**)va;
	const Article* b = *(const Article**)vb;
	return strcmp ( a->message_id, b->message_id );
}

static gint
task_headers_run (Task* task)
{
	StatusItem * status = STATUS_ITEM(task);
	TaskHeaders * task_h = TASK_HEADERS(task);
	GPtrArray * articles = task_h->articles;
	GPtrArray * old_articles = NULL;
	article_db db;
	gboolean active;
	int read;
	int i;
	int worst;
	unsigned long first;
	unsigned long last;
	unsigned long low;
	unsigned long high;
	unsigned long qty;
	const char * progress_fmt;

	if (task_h->download_type == HEADERS_LOCAL) {
		articlelist_load (task->server, task_h->group, status);
		return 0;
	}

	/* find out how many article headers we're getting */
	low = high = first = last = qty = 0;
	progress_fmt = NULL;
	if (task_get_header_range(task_h, &low, &high, &first, &last, &qty, &progress_fmt))
		return 1;

	/* no articles to fetch */
	if (low >= high)
		return 0;


	/* init the progress steps */
	active = task->server==articlelist_get_current_server() &&
	         task_h->group==articlelist_get_current_group();
	i = group_get_attrib_i (task->server, task_h->group, "Total");
	worst = high-low;
	if (TASK_HEADERS(task)->download_type != HEADERS_ALL)
		worst += qty;

	/* get the article headers... */
	status_item_emit_init_steps (status, worst);
	i = nntp_articlelist_download (status,
	                               task->sock,
	                               task->server,
	                               task_h->group,
	                               low,
	                               high,
	                               &task->hint_abort,
	                               TRUE,
	                               progress_fmt,
	                               task_h->articles);
	if (i) /* error; clean up */
	{
		pan_g_ptr_array_foreach (articles, (GFunc)pan_object_unref, 0);
		g_ptr_array_set_size (articles, 0);
		return i;
	}
	group_set_attrib_ul (task->server, task_h->group, "First", first);
	group_set_attrib_ul (task->server, task_h->group, "Last", last);

	status_item_emit_init_steps (status,
		i                         /* load old headers */
		+ (task_h->download_bodies ? worst : 0) /* download bodies */
		+ worst                   /* merge states */
		+ worst                   /* worst-case on save */
		+ (active?(worst*3):0));  /* worse-case on alist_set_contents */

	/* no articles! */
	if (!articles->len)
	{
		status_item_emit_status (status,
			_("No articles retrieved for group %s"),
			task_h->group->name);
		return 0;
	}

	/* bodies */
	if (task_h->download_bodies && task_h->body_index<articles->len)
	{
		int retval = nntp_download_bodies (status,
		                                   task->server,
		                                   task_h->group,
						   task->sock,
						   &task->hint_abort,
						   task_h->articles,
						   &task_h->body_index);
		if (retval)
			return retval;
	}

	log_add_va (
		_("Merging %d article headers with old headers from %s"),
		articles->len,
		task_h->group->name);



	/* merge with existing articles (1) load old articles. */
	db = ahdb_ref (task->server, task_h->group);
	status_item_emit_status (status, _("Loading Old Headers"));
	old_articles = ahdb_get_all (db, status_item_next_step_gfunc, task);
	qsort (old_articles->pdata, old_articles->len, sizeof(gpointer*), pp_article_message_id_comp);

	/* merge with existing articles (2) merge states.
	   This is so that read messages don't suddenly become unread,
	   and so forth. */
	status_item_emit_status (status, _("Merging Headers"));
	for (i=0; i!=articles->len; ++i)
	{
		gboolean exact = FALSE;
		Article* article = ARTICLE(g_ptr_array_index(articles,i));

		const int index = lower_bound (&article,
			old_articles->pdata, old_articles->len,
			sizeof(Article*),
			pp_article_message_id_comp,
			&exact);

		if (exact) { /* got a match! */
			Article* a = ARTICLE(g_ptr_array_index(old_articles,index));
			article->state = a->state;
			pan_object_unref (PAN_OBJECT(a));
			g_ptr_array_remove_index (old_articles, index);
		}
		status_item_emit_next_step (status);
	}

	/* merge with existing articles (3) get list of all existing articles */
	for (i=0; i<old_articles->len; ++i)
	{
		Article* a = ARTICLE(g_ptr_array_index(old_articles,i));
		if (task_h->download_type == HEADERS_ALL) {
			pan_object_unref (PAN_OBJECT(a));
		} else {
			g_ptr_array_add (articles, a);
		}
	}
	g_ptr_array_free (old_articles, TRUE);
	old_articles = NULL;

	/* merge with existing articles (4) update the database */
	read = 0;
	for (i=0; i!=articles->len; ++i) {
		if (article_flag_on (ARTICLE(g_ptr_array_index(articles,i)), STATE_READ))
			++read;
	}
	ahdb_erase_all (db);
	group_set_attrib_i (task->server, task_h->group, "Read", read);
	group_set_attrib_i (task->server, task_h->group, "Total", articles->len);
	status_item_emit_status (status, _("Saving Headers"));
	ahdb_save_all (db, (Article**)articles->pdata, articles->len, status_item_next_step_gfunc, task);

	/* merge with existing articles (5) update UI */
	grouplist_update_row (task_h->group);
	if (active)
	{
		articlelist_set_contents (
			task->server,
			task_h->group,
			(Article**)articles->pdata, articles->len,
			status);
	}

	/* cleanup */
	status_item_emit_progress (status, 0);
	ahdb_sync (db);
	ahdb_unref (db);

	return 0;
}

/*****
******
*****/

static int
task_get_header_range (TaskHeaders  * task,
                       unsigned long * setme_low,
		       unsigned long * setme_high,
		       unsigned long * setme_first,
		       unsigned long * setme_last,
		       unsigned long * setme_qty,
		       const char ** setme_progress_fmt)
{
	const gboolean * const abort = &TASK(task)->hint_abort;
	PanSocket *sock = TASK(task)->sock;
	Server *server = TASK(task)->server;
	Group *group = TASK_HEADERS(task)->group;
	HeaderDownloadType dl_type = TASK_HEADERS(task)->download_type;
	const char * progress_fmt;
	unsigned long total;
	unsigned long first;
	unsigned long last;
	unsigned long low;
	unsigned long high;

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, -1);
	g_return_val_if_fail (server!=NULL, -1);
	g_return_val_if_fail (group!=NULL, -1);

	/* change to the right group */
	total = 0;
	first = 0;
	last = 0;
	if (nntp_get_group_info (STATUS_ITEM(task),
	                         sock,
				 group->name,
				 &total,
				 &first,
				 &last,
				 abort))
		return -1;

	/* figure out which articles to ask for, based on the first/last
	 * returned by the server and the dl_type of this task.
	 */
	if (dl_type==HEADERS_NEW)
	{
		gulong p_last = group_get_attrib_ul (server, group, "Last");
		if (p_last!=0 && p_last==last)
		{
			status_item_emit_status (STATUS_ITEM(task),
				_("No new articles in group '%s'"),
				group->name);
		}

		low = p_last;
		high = last;
		progress_fmt = _("Getting new Article Header %lu of %lu");
	}
	else if (dl_type==HEADERS_SAMPLE)
	{
		low = first;
		if (total>150) {
			status_item_emit_status (STATUS_ITEM(task), _("Getting sample"));
			low = last - 150;
		}
		high = last;
		total = high - low;
		progress_fmt = _("Sampling Article Header %lu of %lu");
	}
	else if (dl_type==HEADERS_ALL)
	{
		low = first;
		high = last;
		progress_fmt = _("Getting Article Header %lu of %lu");
	}
	else
	{
		pan_warn_if_reached ();
		low = 0;
		high = 0;
		progress_fmt = _("You've found the secret message!!");
	}

	*setme_low = low;
	*setme_high = high;
	*setme_first = first;
	*setme_last = last;
	*setme_qty = total;
	*setme_progress_fmt = progress_fmt;
	return 0;
}


static gchar*
task_headers_describe (const StatusItem* item)
{
	const gchar* action = NULL;

	/* sanity checks */
	g_return_val_if_fail (item!=NULL, NULL);

	switch (TASK_HEADERS(item)->download_type)
	{
		case HEADERS_LOCAL: action=_("cached headers"); break;
		case HEADERS_ALL: action=_("all headers"); break;
		case HEADERS_NEW: action=_("new headers"); break;
		case HEADERS_SAMPLE: action=_("latest 150 headers"); break;
		default: pan_warn_if_reached(); action=_("BUG IN CODE"); break;
	}

	return g_strdup_printf ( _("Getting %s for `%s'"),
		action,
		TASK_HEADERS(item)->group->name );
}
