/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <ctype.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-config.h>
#include <libgnomeui/libgnomeui.h>

#include "acache.h"
#include "articlelist.h"
#include "gui.h"
#include "globals.h"
#include "grouplist.h"
#include "message.h"
#include "message-window.h"
#include "pan-object.h"
#include "prefs.h"
#include "queue.h"
#include "task-body.h"
#include "text.h"
#include "util.h"

static GdkColor headers_col = {0, 0xbbbb, 0, 0};

static void ensure_header_col_inited (void);
static void realize_text_cb (GtkWidget *text, gpointer data);
static void text_button_press_cb (GtkWidget *widget, GdkEventButton *bevent);
static void text_key_press_cb (GtkWidget *widget, GdkEventKey *event);
static void text_page_down_or_next_article (GtkText *text);
static void show_all_headers_cb (GtkCheckMenuItem* item, gpointer data);

static GnomeUIInfo text_menu_popup[] =
{
	{
		GNOME_APP_UI_ITEM,
		N_("Open in Window"),
		N_("Open this message for reading."),
		message_read_window
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("_Followup to newsgroup"),
		N_("Post a reply to the message on the news server."),
		message_followup_window, NULL, NULL,
		GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_MAIL_RPL,
		0, 0, NULL
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Reply by _E-mail"),
		N_("Create a mail reply to the sender."),
		message_reply_window, NULL, NULL,
		GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_MAIL_RPL,
		0, 0, NULL
	},
#if 0
	{
		GNOME_APP_UI_ITEM,
		N_("Forward"),
		N_("Forward this message to someone."),
		not_implemented
	},
#endif
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_TOGGLEITEM,
		N_("Show All Headers in Body"),
		N_("Show All Headers in Body"),
		show_all_headers_cb
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("Select All"),
		N_("Select all of the message body"),
		text_select_all,
		NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL,
		'A', GDK_CONTROL_MASK, NULL
	},
	GNOMEUIINFO_END
};

static gchar welcome[] =
{
"Pan " VERSION "\n"
"Copyright (C) 1999, 2000 \n"
"Pan Development Team (pan@superpimp.org) \n"
"\n"
"  http://www.superpimp.org/contact.html   -  Request a Feature\n"
"  http://www.superpimp.org/bugreport.html -  Report a Bug\n"
"  http://www.superpimp.org/download.html  -  Upgrade Pan\n"
"\n"
"This program is free software; you can redistribute it \n"
"and/or modify it under the terms of the GNU General Public \n"
"License as published by the Free Software Foundation; \n"
"either version 2 of the License, or (at your option) any \n"
"later version. \n"
"\n"
"This program is distributed in the hope that it will be \n"
"useful, but WITHOUT ANY WARRANTY; without even the implied \n"
"warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR \n"
"PURPOSE.  See the GNU General Public License for more \n"
"details. \n"
"\n"
"You should have received a copy of the GNU General Public \n"
"License along with this program; if not, write to the \n"
"Free Software Foundation, Inc., 59 Temple Place, \n"
"Suite 330, Boston, MA 02111-1307, USA"
};

static GtkWidget *text_rightclick_popup = NULL;
PanCallback *text_show_all_headers_changed = NULL;
gboolean rot13_enabled = FALSE;
gboolean show_all_headers_in_body = FALSE;

GdkColor text_fg_color;
GdkColor text_bg_color;
GdkColor text_quoted_color;


static void
show_all_headers_cb (GtkCheckMenuItem* item, gpointer data)
{
	text_set_show_all_headers_in_body (item->active);
}

void
text_select_all (void)
{
	pan_lock();
	gtk_editable_select_region (GTK_EDITABLE(Pan.text), 0, -1);
	pan_unlock();
}

void
text_refresh (void)
{
	if (current_article != NULL)
	{
		text_update (current_article);
	}
	else
	{
		char* contents = gtk_editable_get_chars (
			GTK_EDITABLE(Pan.text),0,-1);
		update_body_pane (Pan.text, contents);
		g_free (contents);
	}
}

void
text_set_show_all_headers_in_body (gboolean on)
{
	if (on != show_all_headers_in_body)
	{
		show_all_headers_in_body = on;

		if (current_article != NULL)
		{
			text_update (current_article);
		}

		gtk_check_menu_item_set_active (
			GTK_CHECK_MENU_ITEM(text_menu_popup[5].widget),
			show_all_headers_in_body );
	}

}

/* Scroll down one page, or if we're at the bottom, open the next article */
static void
text_page_down_or_next_article (GtkText *text)
{
        gfloat upper;
	gfloat old_val;

	g_return_if_fail (text != NULL);

	old_val = text->vadj->value;
	text->vadj->value += text->vadj->page_increment;

	upper = text->vadj->upper - text->vadj->page_size;
        text->vadj->value = MIN (text->vadj->value, upper);
        text->vadj->value = MAX (text->vadj->value, 0.0);

	if (text->vadj->value != old_val)
		gtk_signal_emit_by_name (GTK_OBJECT (text->vadj), "value_changed");
	else
		articlelist_read_next_article ();
}

/* Called when the widget is first drawn, puts our welcome text into view */
static void
realize_text_cb (GtkWidget *text, gpointer data)
{
	GtkAdjustment* hadj = GTK_ADJUSTMENT(GTK_TEXT(text)->hadj);
	GtkAdjustment* vadj = GTK_ADJUSTMENT(GTK_TEXT(text)->vadj);

	gui_set_headers_default (NULL, NULL);
	gtk_editable_delete_text (GTK_EDITABLE(text), 0, -1);
	gtk_text_insert (
		GTK_TEXT(text), NULL, &text_fg_color, NULL, welcome, -1);
	gtk_text_set_point (GTK_TEXT(text), 0);
	gtk_adjustment_set_value (GTK_ADJUSTMENT(vadj), 0);
	gtk_text_set_adjustments (GTK_TEXT(text), hadj, vadj);

	gtk_signal_disconnect_by_func (
		GTK_OBJECT(text), GTK_SIGNAL_FUNC(realize_text_cb), NULL);
}

/* Process a mouse button events inside the text widget */
static void
text_button_press_cb (GtkWidget *widget, GdkEventButton *bevent)
{
	gtk_widget_grab_focus (GTK_WIDGET (Pan.text));

	if (bevent->button != 1)
		gtk_signal_emit_stop_by_name (
			GTK_OBJECT(widget), "button_press_event");
	
	if (bevent->button == 3)
	{
		const gboolean sel = current_article != NULL;

		GList* l = GTK_MENU_SHELL(text_rightclick_popup)->children;
		/* l is "open in window" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel);
		l = l->next; /* l is separator */
		l = l->next; /* l is "followup to group */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel);
		l = l->next; /* l is "reply by e-mail" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel);
		l = l->next; /* l is separator */
		l = l->next; /* l is "show all headers in body" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel);
		l = l->next; /* l is separator */
		l = l->next; /* l is "select all" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel);

		gnome_popup_menu_do_popup (text_rightclick_popup, NULL,
					   NULL, bevent, NULL);
	}
}

/* Process a keyboard input event in the text pane */
static void
text_key_press_cb (GtkWidget *widget, GdkEventKey *event)
{
	switch (event->keyval)
	{
		case GDK_KP_Space:
		case GDK_space:
			text_page_down_or_next_article (GTK_TEXT (Pan.text));
			break;
		case GDK_Up:
			if (event->state & GDK_CONTROL_MASK)
				articlelist_read_prev_article ();
			break;
		case GDK_Down:
			if (event->state & GDK_CONTROL_MASK)
				articlelist_read_next_article ();
		default: 
			break;
	}
}

/**
 * text_create:
 * 
 * Return: a GtkScrolledWindow containing the message body text widget
 **/
GtkWidget *
text_create (void)
{
	GtkStyle *style;
	GtkWidget *scrolled_window;

	ensure_header_col_inited();

	Pan.text = gtk_text_new (NULL, NULL);
	gtk_text_set_editable (GTK_TEXT(Pan.text), FALSE);

	widget_set_font (GTK_WIDGET (Pan.text), message_body_font);
	style = gtk_widget_get_style(GTK_WIDGET(Pan.text));
	style->text[0] = text_fg_color;
	style->base[0] = text_bg_color;

#if 0 /* GtkText doesn't allow horizontal scroll.  Rewrite coming in Gtk 1.4 */
	gtk_text_set_word_wrap (GTK_TEXT(Pan.text), FALSE);
	gtk_text_set_line_wrap (GTK_TEXT(Pan.text), FALSE);
#endif

	text_show_all_headers_changed = pan_callback_new ();

	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
	gtk_container_set_border_width (GTK_CONTAINER(scrolled_window), GNOME_PAD_SMALL);

	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_signal_connect (GTK_OBJECT (Pan.text), "realize",
			    GTK_SIGNAL_FUNC(realize_text_cb), NULL);
	gtk_signal_connect (GTK_OBJECT (Pan.text), "button_press_event",
			    GTK_SIGNAL_FUNC(text_button_press_cb), NULL);
	gtk_signal_connect (GTK_OBJECT (Pan.text), "key_press_event",
			    GTK_SIGNAL_FUNC(text_key_press_cb), NULL);

	gtk_container_add (GTK_CONTAINER(scrolled_window), Pan.text);

	text_rightclick_popup = gnome_popup_menu_new (text_menu_popup);

	return scrolled_window;
}

/* Do ROT13 shifting on a message body */
static void
rot13 (gchar* body)
{
	const gchar a [] = "abcdefghijklmnopqrstuvwxyz";
	gchar *pch = NULL;

	for (pch=body; *pch!='\0'; ++pch)
	{
		if (isalpha((int)*pch))
		{
			const gboolean uc = isupper((int)*pch);
			*pch = tolower(*pch);
			*pch = a[((*pch-'a')+13)%26];
			if (uc)
				*pch = toupper(*pch);
		}
	}
}

static void
ensure_header_col_inited (void)
{
	static gboolean headers_col_inited = FALSE;

	if (!headers_col_inited)
	{
		GdkColormap *cmap = NULL;

		pan_lock ();
		cmap = gdk_colormap_get_system ();
		if (gdk_color_alloc (cmap, &headers_col))
			headers_col_inited = TRUE;
		else
			g_error (_("couldn't allocate \"headers\" color"));
		pan_unlock ();
	}
}

void
text_update (Article *article)
{
	gchar *body;

	pan_object_ref (PAN_OBJECT(article));

        /* sanity checks */
        g_return_if_fail (article!=NULL);
        g_return_if_fail (article->message_id!=NULL);
        g_return_if_fail (*article->message_id!='\0');

        /* update active_message_id */
	if (1) {
		Article* tmp = Pan.active_article;
		Pan.active_article = article;
		pan_object_ref (PAN_OBJECT(Pan.active_article));
		if (tmp)
			pan_object_unref (PAN_OBJECT(tmp));
	}
	
        /* load article body */
	body = acache_load_body (article->message_id);

	if (body)
	{
		/* rot13 */
		if (rot13_enabled)
		{
			rot13 (body);
		}

		/* fill */
		if (fill_body_on_show)
		{
			char* pch = fill_body (body);
			g_free (body);
			body = pch;
		}

		/* headers */
		if (show_all_headers_in_body)
		{
			gchar* headers = acache_load_header (article->message_id);
			if (headers)
			{
				gchar* tmp = g_strconcat (headers,"\n\n", body, 0);
				g_free (headers);
				g_free (body);
				headers = NULL;
				body = tmp;
			}
		}

		/* update main text window */
		update_body_pane (Pan.text, body);
		gui_set_headers_default (NULL, article);

		/* update current article */
		if (current_article)
			pan_object_unref (PAN_OBJECT(current_article));
		pan_object_ref (PAN_OBJECT(article));
		current_article = article;

		g_free (body);
	}

	pan_object_unref (PAN_OBJECT(article));
}

void
text_set_from_article (const Server *server,
		       Group *group,
		       Article *article,
		       MessageWindowType type,
		       gboolean needs_window)
{
	/* sanity checks */
	g_return_if_fail (article!=NULL);

	queue_add (TASK(task_body_new(
                     (Server*)articlelist_get_current_server(), //FIXME
                     articlelist_get_current_group(),
                     article,
		     needs_window,
		     type)));
}
